"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var replication_repository_1 = require("../repository/replication-repository");
var system_repository_1 = require("../repository/system-repository");
var _ = require("lodash");
var model_1 = require("../model");
var ReplicationService = (function () {
    function ReplicationService(systemRepository, middlewareClient, replicationRepository) {
        this.systemRepository = systemRepository;
        this.middlewareClient = middlewareClient;
        this.replicationRepository = replicationRepository;
    }
    ReplicationService.getInstance = function () {
        if (!ReplicationService.instance) {
            ReplicationService.instance = new ReplicationService(system_repository_1.SystemRepository.getInstance(), MiddlewareClient.getInstance(), replication_repository_1.ReplicationRepository.getInstance());
        }
        return ReplicationService.instance;
    };
    ReplicationService.prototype.listReplications = function () {
        return this.replicationRepository.listReplications();
    };
    ReplicationService.prototype.listReplicationsForVolume = function (volume) {
        return this.replicationRepository.listReplicationsForVolume(volume);
    };
    ReplicationService.prototype.listReplicationsForDataset = function (dataset) {
        return Promise.all([
            this.replicationRepository.listReplications(),
            this.getHostUuid()
        ]).spread(function (replications, host) { return _.filter(replications, function (replication) {
            return (replication.master === host && replication.datasets[0].master === dataset) ||
                (replication.slave === host && replication.datasets[0].slave === dataset);
        }); });
    };
    ReplicationService.prototype.findOrCreateReplication = function (replicationId) {
        if (replicationId) {
            return this.listReplications().then(function (replications) { return _.find(replications, { id: replicationId }); });
        }
        return this.getNewReplicationInstance();
    };
    ReplicationService.prototype.saveReplication = function (replication) {
        return this.replicationRepository.saveReplication(replication);
    };
    ReplicationService.prototype.syncReplication = function (replicationId) {
        return this.replicationRepository.syncReplication(replicationId);
    };
    ReplicationService.prototype.extractTransportOptions = function (replication) {
        var result = {}, length = replication.transport_options ? replication.transport_options.length || Object.keys(replication.transport_options).length : 0, option, i;
        for (i = 0; i < length; i++) {
            option = replication.transport_options[i];
            if (option['%type'] === 'CompressReplicationTransportOption') {
                result.compress = option.level;
            }
            else if (option['%type'] === 'EncryptReplicationTransportOption') {
                result.encrypt = option.type;
            }
            else if (option['%type'] === 'ThrottleReplicationTransportOption') {
                result.throttle = option.buffer_size;
            }
        }
        return result;
    };
    ReplicationService.prototype.buildTransportOptions = function (options) {
        var promises = [];
        if (options.encrypt) {
            promises.push(this.replicationRepository
                .getNewReplicationTransportOptionInstance(model_1.Model.EncryptReplicationTransportOption)
                .then(function (option) { return _.set(option, 'type', options.encrypt); }));
        }
        if (options.compress) {
            promises.push(this.replicationRepository
                .getNewReplicationTransportOptionInstance(model_1.Model.CompressReplicationTransportOption)
                .then(function (option) { return _.set(option, 'level', options.compress); }));
        }
        if (options.throttle) {
            promises.push(this.replicationRepository
                .getNewReplicationTransportOptionInstance(model_1.Model.ThrottleReplicationTransportOption)
                .then(function (option) { return _.set(option, 'buffer_size', options.throttle); }));
        }
        return Promise.all(promises);
    };
    ReplicationService.prototype.getNewReplicationInstance = function () {
        return Promise.all([
            this.replicationRepository.getNewReplicationInstance(),
            this.getHostUuid()
        ]).spread(function (replication, host) {
            replication.master = host;
            replication.datasets = [{}];
            return replication;
        });
    };
    ReplicationService.prototype.getHostUuid = function () {
        return this.hostPromise = this.hostPromise || this.replicationRepository.getHostUuid();
    };
    return ReplicationService;
}());
exports.ReplicationService = ReplicationService;
