"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var _ = require("lodash");
var crossroads = require("crossroads");
var hasher = require("hasher");
var immutable = require("immutable");
var model_descriptor_service_1 = require("./model-descriptor-service");
var middleware_client_1 = require("./middleware-client");
var model_event_name_1 = require("../model-event-name");
var event_dispatcher_service_1 = require("./event-dispatcher-service");
var datastore_service_1 = require("./datastore-service");
var model_1 = require("../model");
var section_1 = require("../route/section");
var volume_1 = require("../route/volume");
var share_1 = require("../route/share");
var snapshot_1 = require("../route/snapshot");
var dataset_1 = require("../route/dataset");
var calendar_1 = require("../route/calendar");
var system_1 = require("../route/system");
var services_1 = require("../route/services");
var peering_1 = require("../route/peering");
var network_1 = require("../route/network");
var docker_1 = require("../route/docker");
var vms_1 = require("../route/vms");
var accounts_1 = require("../route/accounts");
var replication_1 = require("../route/replication");
var RoutingService = (function () {
    function RoutingService(modelDescriptorService, eventDispatcherService, middlewareClient, datastoreService, sectionRoute, volumeRoute, shareRoute, snapshotRoute, datasetRoute, calendarRoute, systemRoute, serviceRoute, peeringRoute, networkRoute, dockerRoute, replicationRoute) {
        this.modelDescriptorService = modelDescriptorService;
        this.eventDispatcherService = eventDispatcherService;
        this.middlewareClient = middlewareClient;
        this.datastoreService = datastoreService;
        this.sectionRoute = sectionRoute;
        this.volumeRoute = volumeRoute;
        this.shareRoute = shareRoute;
        this.snapshotRoute = snapshotRoute;
        this.datasetRoute = datasetRoute;
        this.calendarRoute = calendarRoute;
        this.systemRoute = systemRoute;
        this.serviceRoute = serviceRoute;
        this.peeringRoute = peeringRoute;
        this.networkRoute = networkRoute;
        this.dockerRoute = dockerRoute;
        this.replicationRoute = replicationRoute;
        this.currentStacks = new Map();
        this.taskStacks = immutable.Map();
        this.sectionRouters = new Map()
            .set('accounts', accounts_1.AccountsRoute.getInstance())
            .set('vms', vms_1.VmsRoute.getInstance());
        this.eventDispatcherService.addEventListener('taskSubmitted', this.handleTaskSubmitted.bind(this));
        this.eventDispatcherService.addEventListener('taskCreated', this.handleTaskCreated.bind(this));
        this.loadRoutes();
        hasher.prependHash = '!';
        hasher.changed.add(this.handleHashChange.bind(this));
    }
    RoutingService.getInstance = function () {
        if (!RoutingService.instance) {
            RoutingService.instance = new RoutingService(model_descriptor_service_1.ModelDescriptorService.getInstance(), event_dispatcher_service_1.EventDispatcherService.getInstance(), middleware_client_1.MiddlewareClient.getInstance(), datastore_service_1.DatastoreService.getInstance(), section_1.SectionRoute.getInstance(), volume_1.VolumeRoute.getInstance(), share_1.ShareRoute.getInstance(), snapshot_1.SnapshotRoute.getInstance(), dataset_1.DatasetRoute.getInstance(), calendar_1.CalendarRoute.getInstance(), system_1.SystemRoute.getInstance(), services_1.ServicesRoute.getInstance(), peering_1.PeeringRoute.getInstance(), network_1.NetworkRoute.getInstance(), docker_1.DockerRoute.getInstance(), replication_1.ReplicationRoute.getInstance());
        }
        return RoutingService.instance;
    };
    RoutingService.prototype.navigate = function (path, force) {
        force = !!force;
        if (hasher.appendHash.length === 0) {
            hasher.appendHash = '?' + location.hash.split('?')[1] || this.middlewareClient.getExplicitHostParam();
        }
        else {
            hasher.appendHash = _.replace(hasher.appendHash, /^\?\?+/, '?');
        }
        if (force) {
            var section = RoutingService.getPathSection(path);
            if (path === '/' + section) {
                this.currentStacks.delete(section);
            }
            else {
                crossroads.ignoreState = true;
                crossroads.parse(path);
                crossroads.ignoreState = false;
            }
        }
        if (path[0] === '/') {
            hasher.setHash(path);
        }
        else {
            hasher.setHash(hasher.getHash() + '/' + path);
        }
    };
    RoutingService.prototype.closeColumn = function (columnIndex) {
        var currentStack = this.getCurrentStack();
        currentStack.splice(columnIndex, currentStack.length - columnIndex);
        this.changeHash(_.last(currentStack).path);
        this.eventDispatcherService.dispatch('pathChange', currentStack);
    };
    RoutingService.prototype.getURLFromObject = function (object) {
        var objectType = this.modelDescriptorService.getObjectType(object), url = objectType === model_1.Model.Section ? '/' : _.kebabCase(objectType), id = !_.isNil(object.id) ?
            object.id :
            !_.isNil(object._tmpId) ?
                object._tmpId :
                null;
        return (Array.isArray(object) || _.isNull(id)) ? url : url + '/_/' + encodeURIComponent(id);
    };
    RoutingService.prototype.handleTaskSubmitted = function (temporaryTaskId) {
        this.saveState(temporaryTaskId);
    };
    RoutingService.prototype.changeHash = function (newHash) {
        hasher.changed.active = false;
        this.navigate(newHash);
        hasher.changed.active = true;
    };
    RoutingService.prototype.handleHashChange = function (newHash, oldHash) {
        var _this = this;
        this.eventDispatcherService.dispatch('inspectorExit').then(function (isBlockingNeeded) {
            if (!isBlockingNeeded) {
                _this.currentSectionId = RoutingService.getPathSection(newHash);
                crossroads.parse(decodeURIComponent(newHash));
                _this.eventDispatcherService.dispatch('hashChange', newHash);
            }
            else {
                _this.changeHash(oldHash);
                _this.eventDispatcherService.dispatch('hashChange', oldHash);
            }
        });
    };
    RoutingService.prototype.getCurrentPath = function () {
        return hasher.getHash();
    };
    RoutingService.prototype.handleTaskCreated = function (taskIds) {
        var _this = this;
        if (this.taskStacks.has(taskIds.old)) {
            var task = this.taskStacks.get(taskIds.old);
            this.taskStacks = this.taskStacks.set(taskIds.new, task).delete(taskIds.old);
            var eventListener_1 = this.eventDispatcherService.addEventListener(model_event_name_1.ModelEventName.Task.change(taskIds.new), function (state) {
                if (state.get('state') === 'FINISHED') {
                    _this.taskStacks = _this.taskStacks.delete(taskIds.new);
                    _this.eventDispatcherService.removeEventListener(model_event_name_1.ModelEventName.Task.change(taskIds.new), eventListener_1);
                }
            });
        }
    };
    RoutingService.prototype.saveState = function (temporaryTaskId) {
        this.taskStacks = this.taskStacks.set(temporaryTaskId, this.getCurrentStateSnapshot());
    };
    RoutingService.prototype.getCurrentStateSnapshot = function () {
        var stateSnapshot;
        if (this.sectionRouters.has(this.currentSectionId)) {
            stateSnapshot = this.sectionRouters.get(this.currentSectionId).saveState();
        }
        else {
            stateSnapshot = [];
            var currentStack = this.currentStacks.get(this.currentSectionId);
            _.forEach(currentStack, function (value, index) {
                var context = _.clone(value);
                if (index > 0) {
                    context.parentcontext = stateSnapshot[index - 1];
                }
                stateSnapshot.push(context);
            });
        }
        return stateSnapshot;
    };
    RoutingService.prototype.getCurrentStack = function () {
        var currentStack;
        if (this.sectionRouters.has(this.currentSectionId)) {
            currentStack = this.sectionRouters.get(this.currentSectionId).getStack();
        }
        else {
            currentStack = this.currentStacks.get(this.currentSectionId);
        }
        return currentStack;
    };
    RoutingService.getPathSection = function (path) {
        return _.head(_.compact(_.split(path, '/')));
    };
    RoutingService.prototype.loadRoutes = function () {
        var _this = this;
        crossroads.addRoute('/_/retry/{taskId}', function (taskId) { return _this.restoreTask(taskId); });
        this.loadDashboardRoutes();
        this.loadStorageRoutes();
        this.loadNetworkRoutes();
        this.loadSettingsRoutes();
        this.loadServicesRoutes();
        this.loadConsoleRoutes();
        this.loadCalendarRoutes();
        this.loadPeeringRoutes();
        this.loadContainersRoutes();
        this.loadWizardRoutes();
    };
    RoutingService.prototype.loadCalendarRoutes = function () {
        var _this = this;
        crossroads.addRoute('/calendar', function () { return _this.calendarRoute.get().then(function (stack) { return _this.currentStacks.set('calendar', stack); }); });
        crossroads.addRoute('/calendar/calendar-task/_/{calendarTaskId}', function (calendarTaskId) { return _this.calendarRoute.getTask(calendarTaskId, _this.currentStacks.get('calendar')); });
        crossroads.addRoute('/calendar/calendar-task/_/{calendarTaskId}/calendar-custom-schedule', function () { return _this.calendarRoute.getCustomSchedule(_this.currentStacks.get('calendar')); });
        crossroads.addRoute('/calendar/calendar-task/create/{taskType}', function (taskType) { return _this.calendarRoute.createTask(taskType, _this.currentStacks.get('calendar')); });
        crossroads.addRoute('/calendar/calendar-task/create/{taskType}/calendar-custom-schedule', function () { return _this.calendarRoute.getCustomSchedule(_this.currentStacks.get('calendar')); });
    };
    RoutingService.prototype.loadNetworkRoutes = function () {
        var _this = this;
        crossroads.addRoute('/network', function () { return _this.loadSection('network'); });
        crossroads.addRoute('/network/settings', function (sectionId) { return _this.sectionRoute.getSettings('network', _this.currentStacks.get('network')); });
        crossroads.addRoute('/network/network-interface/_/{interfaceId}', function (interfaceId) { return _this.networkRoute.get(interfaceId, _this.currentStacks.get('network')); });
        crossroads.addRoute('/network/ipmi', function () { return _this.networkRoute.listIpmi(_this.currentStacks.get('network')); });
        crossroads.addRoute('/network/ipmi/_/{ipmiId}', function (ipmiId) { return _this.networkRoute.getIpmi(ipmiId, _this.currentStacks.get('network')); });
        crossroads.addRoute('/network/create', function () { return _this.networkRoute.selectNewInterfaceType(_this.currentStacks.get('network')); });
        crossroads.addRoute('/network/create/{interfaceType}', function (interfaceType) { return _this.networkRoute.create(interfaceType, _this.currentStacks.get('network')); });
    };
    RoutingService.prototype.loadContainersRoutes = function () {
        var _this = this;
        crossroads.addRoute('/containers', function () { return _this.loadSection('containers'); });
        crossroads.addRoute('/containers/settings', function (sectionId) { return _this.sectionRoute.getSettings('containers', _this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-host', function () { return _this.dockerRoute.listHosts(_this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-host/create', function () { return _this.dockerRoute.createHost(_this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-host/_/{hostId}', function (hostId) { return _this.dockerRoute.getHost(hostId, _this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-host/_/{hostId}/docker-network', function (hostId) { return _this.dockerRoute.listDockerNetworks(hostId, _this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-host/_/{hostId}/docker-network/_/{dockerNetworkId}', function (hostId, dockerNetworkId) { return _this.dockerRoute.getDockerNetwork(dockerNetworkId, _this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-host/_/{hostId}/docker-network/create', function (hostId) { return _this.dockerRoute.createDockerNetwork(hostId, _this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-image', function () { return _this.dockerRoute.listImages(_this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-image/create', function () { return _this.dockerRoute.listCollectionsForCreate(_this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-image/create/{collectionId}', function (collectionId) { return _this.dockerRoute.pullImage(collectionId, _this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-image/_/{imageId}', function (imageId) { return _this.dockerRoute.getImage(imageId, _this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-collection', function () { return _this.dockerRoute.listCollections(_this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-collection/_/{collectionId}', function (collectionId) { return _this.dockerRoute.getCollection(collectionId, _this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-collection/create', function () { return _this.dockerRoute.createCollection(_this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-container', function () { return _this.dockerRoute.listContainers(_this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-container/create', function () { return _this.dockerRoute.listCollectionsForCreate(_this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-container/create/{collectionId}', function (collectionId) { return _this.dockerRoute.createContainer(collectionId, _this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-container/create/{collectionId}/docker-image/{imageName*}/readme', function (containerId, imageName) { return _this.dockerRoute.getReadmeForImage(imageName, _this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-container/_/{containerId}', function (containerId) { return _this.dockerRoute.getContainer(containerId, _this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-container/_/{containerId}/readme', function (containerId) { return _this.dockerRoute.getReadmeForContainer(containerId, _this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/docker-container/_/{containerId}/logs', function (containerId) { return _this.dockerRoute.getContainerLogs(_this.currentStacks.get('containers')); });
        crossroads.addRoute('/containers/section-settings', function () { return _this.dockerRoute.getSettings(); });
    };
    RoutingService.prototype.loadPeeringRoutes = function () {
        var _this = this;
        crossroads.addRoute('/peering', function () { return _this.loadSection('peering'); });
        crossroads.addRoute('/peering/peer/_/{peerId}', function (peerId) { return _this.peeringRoute.get(peerId, _this.currentStacks.get('peering')); });
        crossroads.addRoute('/peering/create', function () { return _this.peeringRoute.selectNewPeerType(_this.currentStacks.get('peering')); });
        crossroads.addRoute('/peering/create/{peerType}', function (peerType) { return _this.peeringRoute.create(peerType, _this.currentStacks.get('peering')); });
    };
    RoutingService.prototype.loadSettingsRoutes = function () {
        var _this = this;
        crossroads.addRoute('/settings', function () { return _this.loadSection('settings'); });
        crossroads.addRoute('/settings/system-section/_/{systemSectionId}', function (systemSectionId) { return _this.systemRoute.get(systemSectionId, _this.currentStacks.get('settings')); });
        crossroads.addRoute('/settings/system-section/_/certificates/crypto-certificate/_/{certificateId}', function (certificateId) { return _this.systemRoute.getCertificate(certificateId, _this.currentStacks.get('settings')); });
        crossroads.addRoute('/settings/system-section/_/certificates/create', function () { return _this.systemRoute.selectNewCertificateType(_this.currentStacks.get('settings')); });
        crossroads.addRoute('/settings/system-section/_/certificates/create/{certificateType}', function (certificateType) { return _this.systemRoute.createCertificate(certificateType, _this.currentStacks.get('settings')); });
        crossroads.addRoute('/settings/system-section/_/alert', function () { return _this.systemRoute.getAlert(_this.currentStacks.get('settings')); }, 1);
        crossroads.addRoute('/settings/system-section/_/alert/settings', function () { return _this.systemRoute.getAlertSettings(_this.currentStacks.get('settings')); });
        crossroads.addRoute('/settings/system-section/_/tunables/tunable/_/{tunableId}', function (tunableId) { return _this.systemRoute.getTunable(tunableId, _this.currentStacks.get('settings')); });
        crossroads.addRoute('/settings/system-section/_/tunables/create', function () { return _this.systemRoute.createTunable(_this.currentStacks.get('settings')); });
        crossroads.addRoute('/settings/system-section/_/ntpservers/ntp-server/_/{ntpServerId}', function (ntpServerId) { return _this.systemRoute.getNtpServer(ntpServerId, _this.currentStacks.get('settings')); });
        crossroads.addRoute('/settings/system-section/_/ntpservers/create', function () { return _this.systemRoute.createNtpServer(_this.currentStacks.get('settings')); });
    };
    RoutingService.prototype.loadServicesRoutes = function () {
        var _this = this;
        crossroads.addRoute('/services', function () { return _this.loadSection('services'); });
        crossroads.addRoute('/services/services-category/_/{categoryId}', function (categoryId) { return _this.serviceRoute.getCategory(categoryId, _this.currentStacks.get('services')); });
        crossroads.addRoute('/services/services-category/_/{categoryId}/service/_/{serviceId}', function (categoryId, serviceId) { return _this.serviceRoute.getService(serviceId, _this.currentStacks.get('services')); });
        crossroads.addRoute('/services/services-category/_/{categoryId}/service/_/{serviceId}/modules', function () { return _this.serviceRoute.listRsyncdModules(_this.currentStacks.get('services')); });
        crossroads.addRoute('/services/services-category/_/{categoryId}/service/_/{serviceId}/modules/create', function () { return _this.serviceRoute.createRsyncdModule(_this.currentStacks.get('services')); });
        crossroads.addRoute('/services/services-category/_/{categoryId}/service/_/{serviceId}/modules/rsyncd-module/_/{moduleId}', function (categoryId, serviceId, moduleId) { return _this.serviceRoute.getRsyncdModule(moduleId, _this.currentStacks.get('services')); });
    };
    RoutingService.prototype.loadDashboardRoutes = function () {
        var _this = this;
        crossroads.addRoute('/dashboard', function () { return _this.sectionRoute.getOld('dashboard'); });
    };
    RoutingService.prototype.loadConsoleRoutes = function () {
        var _this = this;
        crossroads.addRoute('/console', function () { return _this.sectionRoute.getOld('console'); });
    };
    RoutingService.prototype.loadWizardRoutes = function () {
        var _this = this;
        crossroads.addRoute('/wizard', function () { return _this.sectionRoute.getOld('wizard'); });
    };
    RoutingService.prototype.loadStorageRoutes = function () {
        var _this = this;
        crossroads.addRoute('/storage', function () { return _this.loadSection('storage'); });
        crossroads.addRoute('/storage/volume/_/{volumeId}', function (volumeId) { return _this.volumeRoute.get(volumeId, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume/_/{volumeId}/share', function (volumeId) { return _this.shareRoute.list(volumeId, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume/_/{volumeId}/share/create', function (volumeId) { return _this.shareRoute.selectNewType(volumeId, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume/_/{volumeId}/share/create/{type}', function (volumeId, type) { return _this.shareRoute.create(volumeId, type, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume/_/{volumeId}/share/_/{shareId}', function (volumeId, shareId) { return _this.shareRoute.get(volumeId, shareId, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume/_/{volumeId}/volume-snapshot', function (volumeId) { return _this.snapshotRoute.list(volumeId, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume/_/{volumeId}/volume-snapshot/create', function (volumeId) { return _this.snapshotRoute.create(volumeId, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume/_/{volumeId}/volume-snapshot/_/{snapshotId*}', function (volumeId, snapshotId) { return _this.snapshotRoute.get(volumeId, snapshotId, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume/_/{volumeId}/volume-dataset', function (volumeId) { return _this.datasetRoute.list(volumeId, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume/_/{volumeId}/volume-dataset/create', function (volumeId) { return _this.datasetRoute.create(volumeId, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume/_/{volumeId}/volume-dataset/_/{datasetId*}', function (volumeId, datasetId) { return _this.datasetRoute.get(volumeId, datasetId, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume/_/{volumeId}/volume-dataset/_/{datasetId*}/volume-snapshot', function (volumeId, datasetId) { return _this.snapshotRoute.listForDataset(volumeId, datasetId, _this.currentStacks.get('storage')); }, 1);
        crossroads.addRoute('/storage/volume/_/{volumeId}/volume-dataset/_/{datasetId*}/volume-snapshot/create', function (volumeId, datasetId) { return _this.snapshotRoute.createForDataset(volumeId, datasetId, _this.currentStacks.get('storage')); }, 1);
        crossroads.addRoute('/storage/volume/_/{volumeId}/volume-dataset/_/{datasetId*}/volume-snapshot/_/{snapshotId*}', function (volumeId, datasetId, snapshotId) { return _this.snapshotRoute.getForDataset(volumeId, snapshotId, _this.currentStacks.get('storage')); }, 1);
        crossroads.addRoute('/storage/volume/_/{volumeId}/volume-dataset/_/{datasetId*}/vmware-dataset', function (volumeId, datasetId) { return _this.datasetRoute.listVmware(datasetId, _this.currentStacks.get('storage')); }, 1);
        crossroads.addRoute('/storage/volume/_/{volumeId}/volume-dataset/_/{datasetId*}/vmware-dataset/create', function (volumeId, datasetId) { return _this.datasetRoute.createVmware(datasetId, _this.currentStacks.get('storage')); }, 1);
        crossroads.addRoute('/storage/volume/_/{volumeId}/volume-dataset/_/{datasetId*}/vmware-dataset/_/{vmwareDatasetId*}', function (volumeId, datasetId, vmwareDatasetId) { return _this.datasetRoute.getVmware(vmwareDatasetId, _this.currentStacks.get('storage')); }, 1);
        crossroads.addRoute('/storage/volume/_/{volumeId}/volume-dataset/_/{datasetId*}/replication', function (volumeId, datasetId) { return _this.replicationRoute.createForDataset(datasetId, _this.currentStacks.get('storage')); }, 1);
        crossroads.addRoute('/storage/volume/_/{volumeId}/volume-dataset/_/{datasetId*}/share', function (volumeId, datasetId) { return _this.datasetRoute.getShare(volumeId, datasetId, _this.currentStacks.get('storage')); }, 1);
        crossroads.addRoute('/storage/volume/_/{volumeId}/topology', function (volumeId) { return _this.volumeRoute.topology(volumeId, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume/_/{volumeId}/topology/disk/_/{diskId}', function (volumeId, diskId) { return _this.volumeRoute.topologyDisk(volumeId, diskId, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/create', function () { return _this.volumeRoute.create(_this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/create/disk/_/{diskId}', function (diskId) { return _this.volumeRoute.creatorDisk(diskId, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume-importer/_/-', function () { return _this.volumeRoute.import(_this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume-media-importer/_/-', function () { return _this.volumeRoute.mediaImport(_this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume-importer/_/-/detached-volume/_/{volumeId}', function (volumeId) { return _this.volumeRoute.getDetachedVolume(volumeId, _this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume-importer/_/-/detached-volume/_/{volumeId}/topology', function (volumeId) { return _this.volumeRoute.getDetachedVolumeTopology(_this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume-importer/_/-/encrypted', function () { return _this.volumeRoute.importEncrypted(_this.currentStacks.get('storage')); });
        crossroads.addRoute('/storage/volume/_/{volumeId}/replication', function (volumeId) { return _this.replicationRoute.list(volumeId, _this.currentStacks.get('storage')); }, 1);
        crossroads.addRoute('/storage/volume/_/{volumeId}/replication/create', function (volumeId) { return _this.replicationRoute.create(volumeId, _this.currentStacks.get('storage')); }, 1);
        crossroads.addRoute('/storage/volume/_/{volumeId}/replication/_/{replicationId*}', function (volumeId, replicationId) { return _this.replicationRoute.get(replicationId, _this.currentStacks.get('storage')); }, 1);
        crossroads.addRoute('/storage/volume/_/{volumeId}/encrypted-volume-actions', function (volumeId) { return _this.volumeRoute.getEncryptedVolumeActions(volumeId, _this.currentStacks.get('storage')); });
    };
    RoutingService.prototype.loadSection = function (sectionId) {
        var _this = this;
        var promise = this.currentSectionId !== sectionId && this.currentStacks.has(sectionId) ?
            this.restorePreviousSectionStack(sectionId) :
            this.loadSectionStack(sectionId);
        this.currentSectionId = sectionId;
        promise.then(function (stack) {
            if (_this.currentSectionId === sectionId) {
                _this.eventDispatcherService.dispatch('sectionChange', stack[0].service);
                _this.eventDispatcherService.dispatch('pathChange', stack);
            }
        });
    };
    RoutingService.prototype.loadSectionStack = function (sectionId) {
        var _this = this;
        return this.sectionRoute.get(sectionId).then(function (stack) {
            _this.currentStacks.set(sectionId, stack);
            return stack;
        });
    };
    RoutingService.prototype.restorePreviousSectionStack = function (sectionId) {
        var _this = this;
        return new Promise(function (resolve) {
            var stack = _this.currentStacks.get(sectionId);
            _this.changeHash(_.last(stack).path);
            hasher.changed.addOnce(_this.handleTaskHashChange.bind(_this));
            resolve(stack);
        });
    };
    RoutingService.prototype.restoreTask = function (taskId) {
        taskId = _.toNumber(taskId);
        if (this.taskStacks.has(taskId)) {
            var stack = _.clone(this.taskStacks.get(taskId)), section = stack[0].object, sectionId = section.id;
            var taskState = this.datastoreService.getState().get(model_1.Model.Task);
            if (taskState &&
                taskState.get(taskId) &&
                taskState.get(taskId).get('error')) {
                _.last(stack).error = taskState.get(taskId).get('error').toJS();
            }
            if (this.currentStacks.has(sectionId)) {
                this.currentStacks.set(sectionId, stack);
            }
            else {
                this.sectionRouters.get(sectionId).restore(stack);
            }
            this.eventDispatcherService.dispatch('sectionRestored', stack);
            this.changeHash(_.last(stack).path);
        }
    };
    return RoutingService;
}());
exports.RoutingService = RoutingService;
